
from collections import defaultdict, Counter
import json
from dubletten_tool.models import Group
from apis_core.apis_entities.models import Person
from apis_core.apis_metainfo.models import Collection
import logging
from dubletten_tool.logger import init_logger

from apis_core.apis_entities.models import Person
from apis_core.apis_labels.models import LabelType, Label
from apis_core.apis_metainfo.models import Collection
from apis_core.apis_vocabularies.models import PersonPersonRelation
from apis_core.apis_relations.models import PersonPerson, AbstractRelation

from dubletten_tool.models import PersonProxy, Group
import json

log = logging.getLogger("DBLogger")


"""
Functions defined here handle the merging of groups into a vorfinalen eintrag (MergeGroup-Class) and the writing of personperson relations in existing vorfinale Einträge from the related Group.

These functions are used: 
- in the management command to create the inital merges (merge_groups command)
- in merge_views.py, in the remergegroupview, that can be called in the dubletten_tool frontend section. The template for the button is member_list.html, the callback function resides in tool_page.html
"""


# Removed user import here

class MergeGroup:
    merged_col, c = Collection.objects.get_or_create(name="Vorfinale Einträge")
    rt_vorfin, c = PersonPersonRelation.objects.get_or_create(
        name="data merged from", name_reverse="merged into")
    lt_group, c = LabelType.objects.get_or_create(
        name="Result of deduplication Group")
    lt_first_name_alt, c = LabelType.objects.get_or_create(
        name="alternativer Vorname")
    lt_name_alt, c = LabelType.objects.get_or_create(
        name="alternativer Nachname")
    lt_sd_alt, c = LabelType.objects.get_or_create(
        name="alternatives Geburtsdatum")
    lt_ed_alt, c = LabelType.objects.get_or_create(
        name="alternatives Sterbedatum")
    group_map = {}

    rels = ["PersonInstitution", "PersonPlace", "PersonEvent",
            "PersonWork", "PersonPerson"]  # Handle seperately:, "PersonPerson"]

    def __init__(self, group, create=False):
        self.group = group
        self.members = [m.person for m in group.members.all()]
        self.meta = {
            "name": self.members[0].name, "first_name": self.members[0].first_name, "gender": self.group.gender}
        self.create = create
        #self.additional_labels = []

    def process_names(self):
        for m in self.members[1:]:
            self.labels[MergeGroup.lt_first_name_alt].append(
                {"label": m.first_name})
            self.labels[MergeGroup.lt_name_alt].append({"label": m.name})

    def process_birth_and_death(self):
        start_dates = set()
        end_dates = set()
        alt_start, alt_end = None, None
        for m in self.members:
            if m.start_date_written:
                start_dates.add(m.start_date_written)
            if m.end_date_written:
                end_dates.add(m.end_date_written)

        start_dates = list(start_dates)
        end_dates = list(end_dates)

        if start_dates:
            if len(start_dates) == 1:
                start_date = start_dates[0]
            else:
                start_date = start_dates[0]
                alt_start = start_dates[1:]

            #  start_date, *alt_start = [start_dates]

        else:
            start_date = None

        if end_dates:
            if len(end_dates) == 1:
                end_date = end_dates[0]
            else:
                end_date = end_dates[0]
                alt_end = end_dates[1:]
        else:
            end_date = None

        self.meta.update({"start_date_written": start_date,
                         "end_date_written": end_date})

        # todo: create labels from alternative start and end date lists
        if alt_start:
            for sd in alt_start:
                self.labels[MergeGroup.lt_sd_alt].append({"label": sd})
        if alt_end:
            for ed in alt_end:
                self.labels[MergeGroup.lt_ed_alt].append({"label": ed})

            #self.additional_labels.append(label_type=MergeGroup.lt_ed_alt, label=ed)

    def process_labels(self):
        labels = defaultdict(list)
        for m in self.members:
            [labels[el.label_type].append({"label": el.label.strip(
            ), "start_date_written": el.start_date_written, "end_date_written": el.end_date_written}) for el in m.label_set.all()]
        #print("labels is", labels, labels.items())

        return labels

    def process_personinstitution(self):
        pass

    def process_all_other_relations(self):
        # todo: alle relations deduplizieren
        for m in self.members:
            for r in MergeGroup.rels:
                #print("r is:", r)
                if r != "PersonPerson":
                    temp = getattr(m, r.lower()+"_set").values()
                    model = AbstractRelation.get_relation_class_of_name(r)
                    # print(temp)
                    for t in temp:
                        t.pop("related_person_id")
                        t.pop("tempentityclass_ptr_id")
                        t.pop("id")
                        t["related_person"] = self.person
                        # print(t)
                        model.objects.get_or_create(**t)
                elif r == "PersonPerson":
                    # for d in ["personA_set", "personB_set"]:
                    #tempA = m.personA_set.values()
                    #tempB = m.personB_set.values()
                    pass

                    # for t in tempA:
                    #     print(t)
                    #     t.pop("related_personA_id")
                    #     t.pop("tempentityclass_ptr_id")
                    #     t.pop("id")
                    #     t["related_personA"] = self.person
                    #     #print(t)
                    #     model.objects.get_or_create(**t)
                    # for t in tempB:
                    #     print(t)
                    #     t.pop("related_personB_id")
                    #     t.pop("tempentityclass_ptr_id")
                    #     t.pop("id")
                    #     t["related_personB"] = self.person
                    #     #print(t)
                    #     model.objects.get_or_create(**t)

    def process_titles(self):
        titels = set()
        for m in self.members:
            [titels.add(t) for t in m.title.all()]

        return titels

    def process_notes(self):
        notes = ""
        for m in self.members:
            if m.notes:
                notes += f"[{m.pk}: {m.notes}]\n"
        return notes

    def process_references(self):
        refs = ""
        for m in self.members:
            if m.references:
                refs += f"[{m.pk}: {m.references}]\n"
        return refs

    def process_uris(self):
        # nicht nötig
        pass

    def process_collections(self):
        # nicht nötig
        pass

    def run_process(self):

        self.labels = self.process_labels()
        self.process_names()
        self.process_birth_and_death()

        titels = self.process_titles()
        notes = self.process_notes()
        refs = self.process_references()
        self.meta.update({"notes": notes, "references": refs})
        per = Person.objects.create(**self.meta)
        per.name = per.name.strip() + " {vorfinal}"
        per.collection.add(MergeGroup.merged_col)
        per.save()
        # MergeGroup.group_map.update({self.group.id:per.id})
        self.group.vorfin = per
        self.group.save()
        print("merged group")
        # create Ampel
        # ampel = Ampel.objects.create(person=per, status="red", note=self.group.note)
        # ampel.save()

        # todo: handle start and end dates of labels as tuples
        #print(self.labels, self.labels.items())
        for key, vals in self.labels.items():
            for v in vals:
                #print(v, type(v))
                try:
                    lab, created = Label.objects.get_or_create(
                        label_type=key, temp_entity=per, **v)
                except Exception as e:
                    print(e)
                    continue
                # per.label_set.add(lab)
        for m in self.members:
            PersonPerson.objects.get_or_create(
                related_personA=per, related_personB=m, relation_type=MergeGroup.rt_vorfin)

        Label.objects.get_or_create(
            label=f"{self.group.name} ({self.group.id})", label_type=MergeGroup.lt_group, temp_entity=per)
        self.person = per
        self.process_all_other_relations()
        ###print("Notes:", notes, "refs:", refs)


class CountPerPer:
    single_count = 0
    member_count = 0


def write_person_person_rels(g):

    model = PersonPerson
    vorfin = g.vorfin  # Person.objects.get(id=MergeGroup.group_map[g.id])
    print("vorfin is", vorfin)
    rt_vorfin, c = PersonPersonRelation.objects.get_or_create(
        name="data merged from", name_reverse="merged into")

    def get_dublette_or_vorfin(per):
        print("called get_dublette_or_vorfin with per = ", per)
        res = None
        if PersonProxy.objects.filter(person=per).exists():
            print("person existed")
            proxy = PersonProxy.objects.get(person=per)

            if proxy.status == "single":
                print("personproxy status was single")
                CountPerPer.single_count += 1
                print("person was single", per)
                res = per
            else:
                print("person status was not single")

                gs = proxy.group_set.all()
                if len(gs) > 1:
                    print(
                        f"personproxy {proxy} {proxy.id} was member in {gs} - this is {len(gs)}")
                # Person.objects.get(id=MergeGroup.group_map[gs[0].id])
                res = gs[0].vorfin

                if not res:
                    res = per
                    print("person was not single but also not found in group")

                else:
                    CountPerPer.member_count += 1
        else:
            print("person was no PersonProxy", per)
            res = per
        print("res is: ", res)
        return res

    for m in g.members.all():
        ppA = PersonPerson.objects.filter(related_personA=m.person)
        ppB = PersonPerson.objects.filter(related_personB=m.person)
        print("member is", m)
        print(f"ppA {ppA}, ppB {ppB}")
        if ppA:
            for el in ppA:
                print("in ppA, el is: ", el)
                if el.relation_type != rt_vorfin:
                    perB = get_dublette_or_vorfin(el.related_personB)
                    print("perB is: ", perB)
                    print("related_personA is ", vorfin)
                    temp = {
                        "relation_type": el.relation_type,
                        "start_date_written": el.start_date_written,
                        "end_date_written": el.end_date_written,
                        "related_personA": vorfin,
                        "related_personB": perB,
                    }
                    model.objects.get_or_create(**temp)
        if ppB:
            for el in ppB:
                print("in ppB, el is: ", el)
                if el.relation_type != rt_vorfin:
                    perA = get_dublette_or_vorfin(el.related_personA)
                    print("perA is: ", perA)
                    print("realted_personB is: ", vorfin)
                    temp = {
                        "relation_type": el.relation_type,
                        "start_date_written": el.start_date_written,
                        "end_date_written": el.end_date_written,
                        "related_personA": perA,
                        "related_personB": vorfin,
                    }
                    model.objects.get_or_create(**temp)


def update_merging_collections():
    """
    This helper function iterates through all person_proxies and orders each person into the singles or dubletten collection. 
    This function can be called in the frontend to update those collections after changes to groups - like removing a member, thus changing it's status. 
    These changes need to be reflected in the collections. 
    """

    print("update collections called")

    Collection.objects.get(name="Dubletten").delete()
    Collection.objects.get(name="Leopolddaten Singles").delete()

    col_hsv = Collection.objects.get(name="Import HSV full 22-6-21")
    col_hzab = Collection.objects.get(name="Import HZAB full 10-3-21")
    col_acc = Collection.objects.get(name="Import ACCESS full 13-10-21")
    col_new, c = Collection.objects.get_or_create(name="Singles (All ACC+Leopold)")
    col_dubl, c = Collection.objects.get_or_create(name="Dubletten")

    def check_if_leopold_single(person):
        if person.status == "single":
            c = person.person.collection.all()
            if col_hsv in c or col_hzab in c:
                person.person.collection.add(col_new)
                person.person.save()

    for pp in PersonProxy.objects.all():
        if pp.status == "candidate":
            pp.person.collection.add(col_dubl)
            pp.person.save()
        else:
            check_if_leopold_single(pp)

    print("update finished")



from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.views.generic import View, TemplateView
from django.utils.decorators import method_decorator
# Create your views here.
from django.template.loader import render_to_string
from django.contrib.auth.decorators import login_required, user_passes_test
#from .grouping_script_1 import run
from .models import Group, PersonProxy, StatusButtonGroup, StatusButtonProxy, StatusButtonGroupType
import json
from .suggestions import get_levensthein, check_all_names, get_name_suggestions
from .forms import GroupForm, PersonProxyForm
import logging 
from .logger import init_logger

from .merge_views import RemergeGroupView

init_logger()

log = logging.getLogger("DBLogger")


def create_new_buttons(group):
         
    for btn in StatusButtonGroupType.objects.all():
        b, c = StatusButtonGroup.objects.get_or_create(kind=btn, related_instance=group)
        b.save()

    group.save()

@method_decorator(login_required, name="dispatch")
class getToolPage(TemplateView):
    template_name = "tool_page.html"

    def get_context_data(self, **kwargs):
        groups = Group.objects.all()
        context = {}

        context["groups"] = groups
        context["Buttons"] = StatusButtonGroupType.objects.all()
        return context


def calculate_person_suggestions(per_id=None, pp_id=False):
    if not pp_id:
        pp = PersonProxy.objects.get(person__id=per_id)
    else:
        pp = PersonProxy.objects.get(id=pp_id)
    if pp.status == "candidate":
        pgroup = pp.group_set.all()[0]
    else:
        pgroup = None
    lev_data = get_levensthein(pp.person)
    l_singles = []
    l_groups = []
    for el in lev_data:
        if el.status == "single":
            l_singles.append(el)
        elif el.status == "candidate":
            group = el.group_set.all()[0]
            l_groups.append(group)
        else:
            continue
    levs = {"groups": list(set(l_groups)), "singles": list(set(l_singles))}

    names = get_name_suggestions(pp.id)
    all_names = PersonProxy.objects.filter(id__in=names)
    names_singles = all_names.filter(status="single")
    in_group = all_names.filter(status="candidate")
    names_groups = {g for p in in_group for g in p.group_set.all()} 
    if pgroup:
        names_groups.discard(pgroup)    

    data = {"pp": pp, "levs":levs, "names_groups":names_groups, "names_singles":names_singles}

    return data


def get_person_suggestions(request, **kwargs):

    if request.method == "GET":
        per_id = kwargs.get("per_id")
        template = "person_suggestions.html"
        
        data = calculate_person_suggestions(per_id=per_id)

        pp = data["pp"]
        levs = data["levs"]
        names_groups = data["names_groups"]
        names_singles = data["names_singles"]
       
        context = {
            "instance": pp,
            "levs": levs,
            "names": {"groups": list(names_groups), "singles":names_singles}
        }
        html = render_to_string(template, context, request)

        return JsonResponse({"html":html})

def get_group_suggestions(request, **kwargs):

    if request.method == "GET":
        g_id = kwargs.get("group_id")
        group = Group.objects.get(id=g_id)
        members = group.members.all()
        names_groups = set()
        names_singles = set()
        levs = {"groups":set(), "singles":set()}

        for m in members:
            data = calculate_person_suggestions(pp_id=m.id)
            names_groups = names_groups.union(set(data["names_groups"]))
            names_singles = names_singles.union(set(data["names_singles"]))
            levs["groups"] = levs["groups"].union(set(data["levs"]["groups"]))
            levs["singles"] = levs["singles"].union(set(data["levs"]["singles"]))


        template = "person_suggestions.html"
        context = {
            "instance": group,
            "levs": levs,
            "names": {"groups": list(names_groups), "singles":names_singles}
        }

        html = render_to_string(template, context, request)

        return JsonResponse({"html":html})


def rename_group(request):
    if request.method == "POST":
      
        data = json.loads(request.POST.get("data"))
        new_name = data.get("new_name")
        if not new_name or new_name in [None, "Null", "none", "None", "null"]:
            new_name = "Unnamed Group"
        g_id = data.get("g_id")
        group = Group.objects.get(id=g_id)
        old_name = Group.name
        group.name = new_name 
        group.save()
        msg = f"Renamed Group {group.id} with name {group.name} from {old_name} to {new_name} with members {[{'proxy_id':g.id, 'person_id':g.person.id, 'status':g.status} for g in group.members.all()]}."
        log.info(f"MSG: {msg}", extra={"user":request.user, "action": "Renamed Group"})

        return JsonResponse({"data":{"group_id": g_id, "msg":msg, "count":group.count, "new_name":group.name}})

class HandleNoteForm(View):

    def get(self, *args, **kwargs):
        template = "note_form.html"
       
        g_id = self.kwargs.get("g_id")
        type = self.kwargs.get("type")


        if type == "group":
            instance = Group.objects.get(id=g_id)
            form = GroupForm(g_id, instance=instance)

        else: 
            instance = PersonProxy.objects.get(person__id=g_id)
            form = PersonProxyForm(g_id, instance=instance)

        html = render_to_string(template, {"form":form, "instance":instance}, self.request)

        return JsonResponse({"html": html, "success":True})

    def post(self, *args, **kwargs):
        g_id = self.kwargs.get("g_id")
        type = self.kwargs.get("type")


        if type == "group":
            instance = Group.objects.get(id=g_id)
            form = GroupForm(g_id, self.request.POST, instance=instance)

        else:
            instance = PersonProxy.objects.get(person__id=g_id)
            form = PersonProxyForm(g_id, self.request.POST, instance=instance)
            

        success = False
        if form.is_valid():
            ent = form.save()
            ent.save()

            success = True
            res_msg = "Saved Note!"
        else:
            res_msg = "Form was invalid"
        
        log.info(f"Created Note for {type} with id {g_id}. Note is: {ent.note}", extra={"user":self.request.user, "action": "Created/edited Note"})

        return JsonResponse({"msg": res_msg, "success":success, "inst_id":g_id, "type":type})

def get_all_notes(request):
    template = "all_notes.html"
    singles = [{"name":s.name, "id":s.person.id, "note":s.note} for s in PersonProxy.objects.filter(status="single") if s.note]
    candidates = [s.group_set.all()[0] for s in PersonProxy.objects.filter(status="candidate") if s.note]
    groups = [g for g in Group.objects.all() if g.note] + candidates

    groups = set(groups)

    groups = [{"name": g.name, "id":g.id, "note":g.note} for g in groups]
    context = {"groups": groups, "members":candidates, "singles":singles}

    html = render_to_string(template, context, request) 

    return JsonResponse({"html":html})



def get_groups(request, **kwargs):

    if request.method == "GET":
        filter = kwargs.get("val")
        filter = filter.replace("__", " ")
        if filter == "get_all_groups":
            filter = None

        d = json.loads(request.GET.get("data"))
        gender = kwargs.get("gender")
        if gender == "Other":
            gender = [None, "third gender"]
        else: 
            gender = [gender.lower()]
            
        context = {}
        if not filter:
            groups = Group.objects.filter(_gender__in=gender)

        else: 
            groups = Group.objects.filter(name__istartswith=filter, _gender__in=gender)

        groups = list(groups)

        if len(groups) > 100:
            test_groups = set(groups)
            for k, v in d.items():
                if v == "true":
                    val = True
                else: 
                    val = False
                temp = [b.related_instance for b in StatusButtonGroup.objects.filter(kind__id=int(k), value=val)]
                test_groups = test_groups.intersection(set(temp))
            res = [g for g in groups if g in test_groups]
        else:
            res = [g for g in groups if g.check_status(d)]
            res = list(res)

        context["groups"] = res           
        count = len(context["groups"])
        context["group_count"] = count

        html = render_to_string("group_list.html", context, request)

        return JsonResponse({"html": html, "group_count": count})


def renderNoteHtml(instance, type, request):
    template = "note_template.html"

    if instance.note:
        note = instance.note
    else: 
        note = ""

    if type == "group":
        name = instance.name
    else: 
        name = f"{instance.person.name}, {instance.person.first_name}"
        instance = instance.person

    context = {"instance": instance, "note":note, "name":name, "type":type}

    html = render_to_string(template, context, request)

    return html



def getNoteJSON(request, **kwargs):

    if request.method == "GET":
        pk = kwargs.get("inst_id")
        type = kwargs.get("type")
        if type == "group":
            instance = Group.objects.get(id=pk)
        else: 
            instance = PersonProxy.objects.get(person__id=pk)
        
        html = renderNoteHtml(instance, type, request)

        return JsonResponse({"html":html, "success":True})

def get_singles(request, **kwargs):
    
    if request.method == "GET":
        text_name = kwargs.get("val_name")
        text_first = kwargs.get("val_first")

        gender = kwargs.get("gender")
        if gender == "Other":
            gender = [None, "third gender"]
        else: 
            gender = [gender.lower()]

        if text_name == "false":
            text_name = False
        else: 
            text_name = text_name.replace("__", " ")

        if text_first == "false":
            text_first = False
        else:
            text_first = text_first.replace("__", " ")

        if not text_name and not text_first:
            singles = [(s.person.id, s.person.name, s.person.first_name, s.person.start_date, s.person.end_date) for s in PersonProxy.objects.filter(status="single", person__gender__in=gender)]
        elif not text_first:
            singles = [(s.person.id, s.person.name, s.person.first_name, s.person.start_date, s.person.end_date) for s in PersonProxy.objects.filter(person__name__istartswith=text_name).filter(status="single", person__gender__in=gender)]
        elif not text_name:
            singles = [(s.person.id, s.person.name, s.person.first_name, s.person.start_date, s.person.end_date) for s in PersonProxy.objects.filter(person__first_name__icontains=text_first).filter(status="single", person__gender__in=gender)]
        else:
            singles = [(s.person.id, s.person.name, s.person.first_name, s.person.start_date, s.person.end_date) for s in PersonProxy.objects.filter(person__first_name__icontains=text_first, person__name__istartswith=text_name).filter(status="single", person__gender__in=gender)]

        return JsonResponse({"singles":singles})

def get_group_ajax(request, **kwargs):

    if request.method == "GET":
        g_id = kwargs.get("g_id")
        group = Group.objects.get(id=g_id)
        members = [(v, [r for r in v.person.personinstitution_set.all().order_by("relation_type__name")]) for v in group.members.all()]
        members.sort(key=lambda x: len(x[1]), reverse=False)
        vorfin = group.vorfin
        if vorfin:
            vorfin_id = vorfin.id
            vorfin_name = str(vorfin)
        else:
            vorfin_id = None
            vorfin_name = None
        context = {
            "group":group,
            "members": members,
            "vorfin": {"id":vorfin_id, "name":vorfin_name},
        }

        html = render_to_string("member_list.html", context, request)

        return JsonResponse({"html":html})
        
def get_single_ajax(request, **kwargs):

    if request.method == "GET":
        s_id = kwargs.get("s_id")
        single = PersonProxy.objects.get(person__id=s_id)
        if not single.status == "single":
            pass
        
        context = {
            "p": single,
            "rels": [r for r in single.person.personinstitution_set.all().order_by("relation_type__name")]
        }

        html = render_to_string("singles_list.html", context, request)
        return JsonResponse({"html":html})

def update_group_checkboxes(request, **kwargs):
    if request.method == "POST":
        template = "group_buttons.html"
        btn_id = kwargs.get("btn_id")
        g_id = kwargs.get("group_id")

        group = Group.objects.get(id=g_id)
        btn = StatusButtonGroup.objects.get(id=btn_id)
        btn.toggle_status()
        html = render_to_string(template, {"group":group}, request)
        return JsonResponse({"html":html})


def create_new_group(request):
    if request.method == "POST":
        d = json.loads(request.POST.get("DATA"))


        data = d.get("data")
        name = d.get("new_name")
    
        if not name:
            name = "New Group"

        new_group = Group.objects.create(name=name)
        if name == "New Group":
            new_group.name += " "+str(new_group.id)
            new_group.save()

        updates = {}
        former_singles = []
        for k, v in data.items():
            if v:
                if k != "singles":
                    group = Group.objects.get(id=k)
                   
                pers = PersonProxy.objects.filter(person__id__in=v)
                for p in pers:
                    if k != "singles":
                        group.members.remove(p)
                        group.save()
                    else: 
                        p.status = "candidate"
                        former_singles.append(p.person.id)
                        p.save()
                    new_group.members.add(p)

                if k != "singles":
                    count = group.count
                    g_id = group.id
                    if count == 0:
                        log.info(f"Deleted Group {group.id} with name {group.name} because group had no members", extra={"user":request.user, "action": "Deleted Group"})

                        group.delete()
                        

                    elif count == 1:
                        per = group.members.all()[0]
                        per.status = "single"
                        per.save()
                        log.info(f"Deleted Group {group.id} with name {group.name} because only one member left {per.id} with name {per.name} and status {per.status}", extra={"user":request.user, "action": "Deleted Group"})

                        group.delete()
                        count = 0

                    else:
                        group.save()
               
                    updates.update({g_id:count})
                new_group._gender = new_group.members.all()[0].person.gender
                new_group.save()
        create_new_buttons(new_group)

        res = {"new_group_id": new_group.id, "new_group": new_group.name, "new_group_count":new_group.count, "former_singles":former_singles,"group_updates":updates}
        
        log.info(f"Created new group {new_group.id}, with name {new_group.name}, with members {[{'proxy_id':g.id, 'person_id':g.person.id, 'status':g.status} for g in new_group.members.all()]}, with former_singles {former_singles} and group updates {updates}", extra={"user":request.user, "action": "Created New Group"})
        return JsonResponse({"data":res})




def merge_groups(request):
    if request.method == "POST":
        d = json.loads(request.POST.get("DATA"))
        name = d.get("new_name")
        groups = d.get("groups")
        singles = d.get("singles")

        new_group = Group.objects.create(name=name)
        log_groups = []

        old_vorfins = []

        def get_vorfin_of_group(group):
            if group.vorfin: 
                return group.vorfin
            else: 
                print(f"ADDITONAL LOGIC: Group {group} had no vorfin. Returnin None")
                return None

        if groups:
            for g in groups:
                # fetch vorfins here
                # send vorfin rels 
               
                group = Group.objects.get(id=g)

                old_vorfin = get_vorfin_of_group(group)
                if old_vorfin: 
                    old_vorfins.append(old_vorfin)

                log_groups.append((group.id, group.name))
                for m in group.members.all():
                    new_group.members.add(m)
                note = group.note
                if note and new_group.note:
                    new_group.note += f"\n[[ Notes from merged group ({g}):{note}]]"
                elif note:
                    new_group.note = f"[[ Notes from merged group ({g}):{note}]]"

                new_group.save()
                log.info(f"Deleted Group {group.id} with name {group.name} because group was merged into {new_group.id} with name {new_group.name}", extra={"user":request.user, "action": "Deleted Group"})
                group.delete()

        if singles: 
            for s in singles:
                pp = PersonProxy.objects.get(person__id=s)
                pp.status = "candidate"
                pp.save()
                new_group.members.add(pp)
            new_group.save()
        
        new_group._gender = new_group.members.all()[0].person.gender
        new_group.save()
        create_new_buttons(new_group)


        remerge_data = RemergeGroupView.get_remerge_data(new_group, vorfin_list=old_vorfins)
        log.info(f"Merged Groups - groups {log_groups} and singles: {singles} into new group {new_group.id}, with name {new_group.name} and members {[{'proxy_id':g.id, 'person_id':g.person.id, 'status':g.status} for g in new_group.members.all()]}", extra={"user":request.user, "action": "Merged Groups"})
        
        response = {
        "group_data" : {"remove_groups": groups, "remove_singles":singles, "add": [new_group.id, new_group.name, new_group.count]},
        "remerge_data" : remerge_data,
        }
        return JsonResponse(response)
            



def remove_member(request, **kwargs):
    if request.method == "GET":
        group_id = kwargs.get("group_id")
        per_id = kwargs.get("per_id")

        group = Group.objects.get(id=group_id)
        per = PersonProxy.objects.get(person__id=per_id)
        group.members.remove(per)
        per.status = "single"
        group.save()
        per.save()
        updates = {}
        log_group = f"group {group.id} with name {group.name} and members {[{'proxy_id':g.id, 'person_id':g.person.id, 'status':g.status} for g in group.members.all()]}"

        count = group.count
        if count == 0:
            log.info(f"Deleted Group {group.id} with name {group.name} because group had no members", extra={"user":request.user, "action": "Deleted Group"})

            group.delete()
            

        elif count == 1:
            per = group.members.all()[0]
            per.status = "single"
            per.save()
            log.info(f"Deleted Group {group.id} with name {group.name} because group had only one member left {per.id} with name {per.name} now status {per.status}", extra={"user":request.user, "action": "Deleted Group"})
            group.delete()
            count = 0

        else:
            group.save()
        
        updates.update({group_id:count})

        log.info(f"Removed {per.id} with status (now) {per.status} from DATA before removing person: {log_group}", extra={"user":request.user, "action": "Removed Person from Group"})
     

        return JsonResponse({"data":updates})

            





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dubletten_tool',
 'dubletten_tool.dump_improvements.helper_functions',
 'dubletten_tool.management',
 'dubletten_tool.management.commands',
 'dubletten_tool.migrations',
 'dubletten_tool.src.classes',
 'dubletten_tool.src.functions',
 'dubletten_tool.tests']

package_data = \
{'': ['*'],
 'dubletten_tool': ['dump_improvements/*',
                    'static/dubletten_tool/css/*',
                    'templates/*',
                    'templates/del/*',
                    'templates/merge/*',
                    'templates/merge/detail_views/*']}

install_requires = \
['python-Levenshtein>=0.12.2,<0.13.0']

setup_kwargs = {
    'name': 'dubletten-tool',
    'version': '0.2.0',
    'description': 'APIS module for the VieCPro instance to perform manual deduplication of person instances',
    'long_description': 'None',
    'author': 'Gregor Pirgie',
    'author_email': 'gregor.pirgie@oeaw.ac.at',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import logging
from .models import DublettenLOG
from datetime import datetime

class DatabaseHandler(logging.Handler):
    """
    Custom Handler that receives logging records and writes it to the Database. 

    Context information for the 'user' and 'action' fields is added in the log argument "extra" which expects a dictionary.
    This context information is added to the records '__dict__' object, and can be accessed by the keys given in the "extra" argument 
    provided in the call to log.info, etc. 
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    
    def emit(self, record):
        user = record.__dict__.get("user")
        action = record.__dict__.get("action")
        message = self.format(record)
        DublettenLOG.objects.create(msg=message, user=user, action=action)


def init_logger():
    LOG = logging.getLogger("DBLogger")
    #format = logging.Formatter("<<<<< LOGGER >>>>> %(module)s >>> %(funcName)s line:%(lineno)d >>> %(message)s")
    db_format = logging.Formatter("%(user)s >>> %(action)s >>> %(funcName)s >>> %(message)s")
    #handler = logging.StreamHandler()
    #file = logging.FileHandler()
    #handler.setFormatter(format)
    handlerDB = DatabaseHandler()
    handlerDB.setFormatter(db_format)
    LOG.addHandler(handlerDB)
    #LOG.addHandler(handler)
    LOG.setLevel(logging.INFO)






# Generated by Django 3.1.14 on 2022-06-01 14:14

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dubletten_tool', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DublettenLOG',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('msg', models.TextField(max_length=600)),
                ('action', models.CharField(max_length=255)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 3.2 on 2022-11-22 08:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dubletten_tool', '0005_alter_ampel_status'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Ampel',
        ),
    ]



# Generated by Django 3.2 on 2022-10-26 18:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('apis_entities', '0004_auto_20200722_1231'),
        ('dubletten_tool', '0003_ampel'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='ampel',
            name='id',
        ),
        migrations.AddField(
            model_name='ampel',
            name='status',
            field=models.CharField(choices=[('red', 'red'), ('yellow', 'yelow'), ('green', 'green')], default='red', max_length=300),
        ),
        migrations.AlterField(
            model_name='ampel',
            name='person',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='apis_entities.person'),
        ),
    ]
